import wave
import os

import numpy as np

from stegosphere.containers import container

class WAVContainer(container.Container):
    """
    File container for wave files, for reading, writing and saving.
    """
    def __init__(self, audio):
        if isinstance(audio, wave.Wave_read):
            self.audio = audio
        elif os.path.isfile(audio):
            self.audio = wave.open(audio, 'rb')
        elif isinstance(audio, np.ndarray):
            raise Exception('No need for containers for arrays')

        else:
            raise Exception('wav must be wave.Wave_read or path')

        self.meta = self.audio.getparams()
        self.frames = None
    def read(self):
        """
        Read frames.
        """
        frame_bytes = self.audio.readframes(self.audio.getnframes())
        sample_width = self.audio.getsampwidth()
        dtype = np.int16 if sample_width == 2 else np.uint8  # Adjust for bit depth
        samples = np.frombuffer(frame_bytes, dtype=dtype)
        self.audio.setpos(0)
        return samples.reshape(-1, self.meta.nchannels)

    def flush(self, frames):
        """
        Flush new frames into container.

        :param frames:
        :type frames: numpy.ndarray
        """
        self.frames = frames.flatten().tobytes()

    def save(self, path, keep_meta=True):
        """
        Save audio to path.

        :param path: Path to new file
        :param keep_meta: Keep basic information from initial file
        """
        assert self.frames is not None, 'Flush frames first'
        with wave.open(path, 'wb') as output:
            if keep_meta:
                output.setparams(self.meta)
            output.writeframes(self.frames)




#The standard End-of-message signifier - gets converted to binary. Long enough for all standard use cases.
DELIMITER_MESSAGE = '###END###'


#32 bit for metadata length should be sufficient for almost all usecases
#Set higher if hidden message exceeds ~0.5GB
#Set lower if hidden message needs to be minimized
METADATA_LENGTH_IMAGE = 32
METADATA_LENGTH_AUDIO = 32
METADATA_LENGTH_VIDEO = 32
METADATA_LENGTH_TTF = 32

METADATA_LENGTH_LSB = 32
METADATA_LENGTH_VD = 32



from numpy import random
import numpy as np

from stegosphere import utils, io

__all__ = ['roundrobin_chunks', 'split_encode','split_decode', 'weighted_chunks']

def weighted_chunks(payload, num_instances, weights):
    """
    Splits payload into num_instances chunks according to weights.
    """
    assert sum(weights)==1
    assert len(weights)==num_instances

    length = len(payload)
    chunk_sizes = [int(round(w * length)) for w in weights]
    
    #fix rounding at last chunk
    diff = length - sum(chunk_sizes)
    if diff > 0:
        chunk_sizes[-1] += diff
    elif diff < 0:
        chunk_sizes[-1] += diff
    
    chunks = []
    start = 0
    for size in chunk_sizes:
        end = start + size
        chunks.append(payload[start:end])
        start = end
        
    return chunks


def roundrobin_chunks(payload, num_instances):
    """
    Distribute payload in a cycle / round-robin.
    """
    rosters = [[] for _ in range(num_instances)]
    for i, bit in enumerate(payload):
        rosters[i % num_instances].append(bit)

    return [''.join(chunk) for chunk in rosters]

def reverse_roundrobin(payload, num_instances):
    """
    Reconstruct payload from round-robin payload.
    
    :param payload: The entire string distributed using round-robin.
    :type payload: str
    :param num_instances: The number of instances used
    :type num_instances: int
    
    :return: Reconstructed payload.
    :rtype: str
    """
    chunk_lengths = [len(payload) // num_instances] * num_instances
    remainder = len(payload) % num_instances
    
    for i in range(remainder):
        chunk_lengths[i] += 1
    
    chunks = []
    start = 0
    for length in chunk_lengths:
        chunks.append(payload[start:start + length])
        start += length
    

    output = [''] * len(payload)
    positions = [0] * num_instances
    for i in range(len(payload)):
        chunk_index = i % num_instances
        output[i] = chunks[chunk_index][positions[chunk_index]]
        positions[chunk_index] += 1
    
    return ''.join(output)




def split_encode(payload, instances, seed=None, distribution='even', distribution_args=None):
    """
    Encodes a payload across several instances.

    :param payload: The payload to be encoded.
    :type payload: str
    :param instances: The iterable of instances.
    :type instances: list
    :param seed: Seed to pseudo-randomly distribute the payload over the different instances.
    :type seed: int, optional
    :param distribution: How to distribute the payload across the instances. Defaults to even distribution.
    :type distribution: str, optional
    :param distribution_args: Additional distribution args
    :type distribution_args: dict, optional

    :return: outputs
    :rtype: list
    """

    
    if seed:
        indices = utils.prng_indices(len(payload), seed)
        payload = ''.join(payload[i] for i in indices)
    else:
        indices = np.arange(len(payload))

    num_instances = len(instances)
    payload_chunks = []
    output = []
    
    if distribution == 'even':
        chunk_length = len(payload)//num_instances
        for i in range(len(instances)):
            payload_chunks.append(payload[i*chunk_length:(i+1)*chunk_length])
        if remainder := len(payload) % len(instances):
            payload_chunks[-1] += payload[-remainder:]
    elif distribution == 'weighted':
        weights = distribution_args.get('weights')
        payload_chunks = weighted_chunks(payload, num_instances, weights)
    elif distribution == 'roundrobin':
        payload_chunks = roundrobin_chunks(payload, num_instances)
    else:
        raise NotImplementedError(f"Distribution {distribution} unknown")

    for instance, chunk in zip(instances, payload_chunks):
        output.append(instance(chunk))
    return output

    
def split_decode(instances, seed=None, distribution=None, distribution_args=None):
    """Decodes a payload across several instances.

       :param instances: The iterable of instances.
       :type instances: list
       :param seed: Seed to pseudo-randomly distribute the payload over the different instances.
       :type seed: int, optional
       :param distribution: How to distribute the payload across the instances.
       :type distribution: str, optional
       :param distribution_args: Additional distribution args
       :type distribution_args: dict, optional

       :return: The decoded payload
       :rtype: str
    """
    output = ''
    for instance in instances:
        output += instance()
    if distribution == 'roundrobin':
        output = reverse_roundrobin(output, len(instances))
        
    if seed is None:
        return output
    else:
        indices = utils.prng_indices(len(output),seed)
        zeros = np.zeros(len(output),dtype=int)
        for bit, index in zip(output, indices):
            zeros[index] = bit
        return ''.join(map(str, zeros))



import numpy as np

def hamming_distance(before, after):
    """
    Compute Hamming Distance (HD)

    :param before:
    :param after:

    :return: Hamming Distance
    """
    assert len(before) == len(after)
    return np.sum(np.array(before) != np.array(after))

def bit_error_rate(before, after):
    """
    Compute Bit Error Rate (BER)
    """
    assert len(before) == len(after)
    if before == '':
        return 0
    return hamming_distance(before, after) / len(before)

def normalized_correlation(before, after):
    """
    Compute Normalized Correlation (NC)
    """
    assert len(before) == len(after)
    before = np.array(before)
    after = np.array(after)
    numerator = np.sum(before * after)
    denominator = np.sum(before ** 2)
    return numerator / denominator if denominator != 0 else 0



from stegosphere.io import binary_to_data, binary_to_file, data_to_binary, file_to_binary

__all__ = ['binary_to_data', 'binary_to_file', 'data_to_binary', 'file_to_binary']

__author__ = 'Maximilian J. W. Koch'

__version__ = '1.2.1'



import re

from stegosphere.tools import compression

def encode_payload(payload, method='metadata', metadata_length=32, delimiter_message='###END###', compress=False):
    """
    Prepares a payload for embedding.

    :param payload: The payload to be embedded
    :type payload: str
    :param method: The method for end-of-message signifying, either 'metadata', 'delimiter' or None
    :type method: str/None
    :param metadata_length: The length of the block denoting the number of embedded bits.
    :type metadata_length: int
    :param delimiter_message: The delimiter to use as end-of-message
    :type delimiter_message: str
    :param compress: Use compression to compress the payload
    :type compress: bool, str
    """
    payload_format = check_type(payload)
    if payload_format == 1:
        #Hex messages are a common output after encryption, thus treated specifically
        payload = hex_to_binary(payload)
    elif payload_format == 2:
        payload = data_to_binary(payload)

    if compress:
        previous_length = len(payload)
        payload = compression.binary_compress(payload, compress)
        after_length = len(payload)
        if after_length > previous_length:
            warnings.warn('message length increased due to compression. That might be the case for already compressed data.')
        
    if method == 'delimiter':
        if not is_binary(delimiter_message):
            delimiter_message = data_to_binary(delimiter_message)
        payload += delimiter_message
        assert payload.count(delimiter_message)==1, Exception('Delimiter appears in data. Use another delimiter or change data minimally.')
    elif method == 'metadata':
        payload = f"{len(payload):0{metadata_length}b}" + payload
    elif method is not None:
        raise Exception('Method must be either delimiter, metadata or None.')
    
    return payload


def file_to_binary(path):
    """
    Converts a file into its binary representation.
    
    :param file_path: Path to the input file
    :return: Binary data of the file
    """
    with open(path, 'rb') as file:
        binary_data = file.read()
    return data_to_binary(binary_data)

def binary_to_file(binary_data, output_path):
    """
    Converts binary data back to a file.
    
    :param binary_data: Binary data to convert
    :param output_file_path: Path to save the output file
    """
    data = binary_to_data(binary_data)
    with open(output_path, 'wb') as file:
        file.write(data)

def data_to_binary(data):
    """
    Converts data (string or bytes) to a binary string.
    
    :param data: Data to convert
    :return: Binary string
    """
    if isinstance(data, str):
        return ''.join(format(ord(char), '08b') for char in data)
    elif isinstance(data, bytes) or isinstance(data, np.ndarray):
        return ''.join(format(byte, '08b') for byte in data)
    elif isinstance(data, int) or isinstance(data, np.uint8):
        return format(data, '08b')
    else:
        raise TypeError("Type not supported.")

def binary_to_data(binary):
    """
    Converts binary string to data.
    
    :param binary: Binary string to convert
    :return: Data as bytes
    """
    all_bytes = [binary[i: i + 8] for i in range(0, len(binary), 8)]
    return bytes([int(byte, 2) for byte in all_bytes])

def is_binary(data):
    """
    Check if string only contains binary data.

    :param data: data to be checked
    :return: bool
    """
    return re.fullmatch(r'[01]+', data)


def hex_to_binary(data):
    """
    Convert hexadecimal str to binary str.

    :param data: data to be converted
    :return: binary data
    """
    return bin(int(data, 16))[2:]

def check_type(data):
    """
    Check whether data is binary, hexadecimal or neither

    :param data: data to be checked
    :return: int, 0 meaning binary, 1 meaning hex, 2 meaning neither
    """
    if is_binary(data):
        return 0 #binary
    elif re.fullmatch(r'[0-9a-fA-F]+', data):
        return 1 #hexadecimal
    else:
        return 2 #to be treated as string




import numpy as np


"""
payload_size: size of payload in bits embedded in the array
max_capacity: maximum size of payload that can be embedded
"""

def max_capacity_estimator(array, method):
    pass

def bits_per_value(array, max_capacity):
    return max_capacity / array.size

def payload_ratio(payload_size, max_capacity):
    return payload_size / max_capacity

def bits_per_value_used(array, payload_size):
    return payload_size / array.size



import os
import numpy as np
try:
    import PIL.Image
except:
    raise Exception('ImageContainer requires PIL. Use array container instead.')

from stegosphere.containers import container

class ImageContainer(container.Container):
    """
    File container for image files.
    """
    def __init__(self, image):
        if isinstance(image, PIL.Image.Image):
            self.image = image
        elif os.path.isfile(image):
            self.image = PIL.Image.open(image)
        elif isinstance(image, np.ndarray):
            raise Exception('No need for containers for arrays')
        else:
            raise Exception('image must be PIL.Image or path')
        
    def read(self):
        """
        Read pixels
        """
        return np.array(self.image)
    def flush(self, pixels):
        """
        Flush pixels into container
        """
        assert isinstance(pixels, np.ndarray)
        if len(pixels.shape) <= 2: #grayscale
            data = pixels.flatten()
        else: #multiple channels
            data = [tuple(pixel) for row in pixels for pixel in row]
        self.image.putdata(data)
    def save(self, path):
        """
        Save image to path
        """
        self.image.save(path)
            



#Measures time and memory usage for efficiency

import timeit
import tracemalloc

class Timer:
    def __enter__(self):
        self.start = timeit.default_timer()
        return self

    def __exit__(self, *args):
        self.end = timeit.default_timer()
        self.diff = self.end - self.start 



import numpy as np


#include option for looping on RGB(A) images!

def transform(array, loop_3d=False):
    """
    Transform data into wavelet domain.
    """
    data, adjusted = _adjust_for_uneven_lengths(array)
    return _iwt_nd(data), adjusted
    
def inverse(array, adjusted):
    """
    Inverse transform data into spatial domain.
    """
    data = _iiwt_nd(array)
    return _restore_uneven_lengths(data, adjusted)


def _adjust_for_uneven_lengths(array):
    """
    Removes last element in uneven dimensions.

    Return array with even lengths and dictionary with removed element for each dimension.
    """
    adjusted_array = array.copy()
    removed_elements = {}

    for axis in reversed(range(array.ndim)):
        dim_size = adjusted_array.shape[axis]
        if dim_size % 2 == 0:
            continue
        else:
            slices = [slice(None)] * adjusted_array.ndim
            slices[axis] = slice(0, -1)
            
            last_element_slices = [slice(None)] * adjusted_array.ndim
            last_element_slices[axis] = slice(-1, None)
            removed_elements[axis] = adjusted_array[tuple(last_element_slices)]
            
            adjusted_array = adjusted_array[tuple(slices)]
    
    return adjusted_array, removed_elements

def _restore_uneven_lengths(array, removed_elements):
    """
    Restore adjusted array.
    """
    restored_array = array.copy()
    for axis in sorted(removed_elements.keys()):
        elements = removed_elements[axis]
        restored_array = np.concatenate([restored_array, elements], axis=axis)
    return restored_array


def _iwt_nd(array):
    """
    Perform an n-dimensional integer Haar wavelet transform.

    :param array: n-dimensional integer array
    :type array: np.ndarray

    :return: Dictionary with coefficients. Keys are tuples, with 0 for approximation and 1 for detail.
    :rtype: dict
    """
    array = array.astype(np.int64)

    if any(dim % 2 != 0 for dim in array.shape):
        raise ValueError("All dimensions of the input array must be even.")

    coeffs = {(): array}

    for axis in range(array.ndim):
        new_coeffs = {}
        for key, arr in coeffs.items():
            s = [slice(None)] * arr.ndim

            s_even = s.copy()
            s_even[axis] = slice(None, None, 2)
            arr_even = arr[tuple(s_even)]

            s_odd = s.copy()
            s_odd[axis] = slice(1, None, 2)
            arr_odd = arr[tuple(s_odd)]

            approx = (arr_even + arr_odd) // 2
            detail = arr_even - arr_odd

            key_approx = key + ('0',)
            key_detail = key + ('1',)

            new_coeffs[key_approx] = approx
            new_coeffs[key_detail] = detail

        coeffs = new_coeffs

    return coeffs

def _iiwt_nd(coeffs):
    """
    Perform the inverse n-dimensional integer Haar wavelet transform.

    :param coeffs: Dictionary with coefficients.
    :param type: dict

    :return: Reconstructed array.
    :rtype: np.ndarray
    """
    ndim = len(next(iter(coeffs.keys())))  # Number of dimensions
    coeffs_inv = coeffs.copy()

    for axis in reversed(range(ndim)):
        new_coeffs = {}
        keys = set(k[:-1] for k in coeffs_inv.keys())

        for key in keys:
            key_approx = key + ('0',)
            key_detail = key + ('1',)

            approx = coeffs_inv[key_approx]
            detail = coeffs_inv[key_detail]

            s = [slice(None)] * approx.ndim
            shape = list(approx.shape)
            shape[axis] *= 2
            arr = np.zeros(shape, dtype=np.int64)

            s_even = s.copy()
            s_even[axis] = slice(None, None, 2)
            s_odd = s.copy()
            s_odd[axis] = slice(1, None, 2)

            arr_even = approx + ((detail + 1) // 2)
            arr_odd = arr_even - detail

            arr[tuple(s_even)] = arr_even
            arr[tuple(s_odd)] = arr_odd

            new_coeffs[key] = arr

        coeffs_inv = new_coeffs
    return coeffs_inv[()]



import numpy as np

def dtype_range(dtype):
    """
    Gets minimum/maximum value of numpy datatype.

    :param dtype: dtype to be checked
    :return: tuple, with minimum and maximum value.
    """
    if np.issubdtype(dtype, np.integer):
        info = np.iinfo(dtype)
    elif np.issubdtype(dtype, np.floating):
        info = np.finfo(dtype)
    else:
        raise Exception('array dtype must be integer or float.')
    return info.min, info.max

def prng_indices(length, key):
    """
    Shuffles indices of array using PCG64DXSM.

    :param length: Length of array.
    :param key: Seed to use for the PRNG.
    :return: Shuffled indices.
    """
    if type(key)!=int:
        key = np.frombuffer(key.encode(), dtype=np.uint32)
    rng = np.random.Generator(np.random.PCG64DXSM(seed=key))
    indices = np.arange(length)
    rng.shuffle(indices)
    return indices

def generate_binary_payload(length):
    """
    Generate a binary payload.

    :param length: The length of the payload.
    :type length: int

    :return: payload
    :rtype: str
    """
    binary_array = np.random.randint(0, 2, size=length)
    return ''.join(map(str, binary_array))



#robustness attacks

import numpy as np


def random_bit_attack(array, ratio=0.01):
    """
    Change given ratio of array by +1 or -1, randomly.
    """
    array = array.copy()
    indices = np.random.choice(array.size, int(array.size * ratio), replace=False)
    array.flat[indices] += np.random.choice([-1, 1], size=len(indices))
    return array



#Detects the likelihood of steganography

import numpy as np
import random

def classifier_evaluation(detector, inputs, outputs):
    """
    Evaluates a steganalysis classifier.

    :param detector: detector function
    :type detector: callable
    :param inputs: inputs to classify
    :type inputs: iterable
    :param outputs: true outputs
    :type outputs: iterable
    """
    
    assert len(inputs)==len(outputs)
    t_p, t_n, f_p, f_n = 0,0,0,0
    for inp, out in zip(inputs, outputs):
        estimate = detector(inp)
        if estimate == 1 and out == 1: t_p += 1
        elif estimate == 1 and out == 0: f_p += 1
        elif estimate == 0 and out == 1: f_n += 1
        elif estimate == 0 and out == 0: t_n += 1

    return {'true_positive' : t_p, 'true_negative' : t_n,
            'false_positive' : f_p, 'false_negative' : f_n}

#For test purposes
def random_detector(input):
    return random.randint(0,1)


def uniformity_detector(array):
    """
    (for test purposes)
    Detects the likelihood of LSB steganography in an RGB image.
    """
    if len(array.shape) != 3 or array.shape[2] != 3:
        raise Exception("Input must be RGB image array.")
    
    #LSB of each channel
    lsb_red = array[:, :, 0] % 2
    lsb_green = array[:, :, 1] % 2
    lsb_blue = array[:, :, 2] % 2
    lsb_all = np.concatenate([lsb_red.flatten(), lsb_green.flatten(), lsb_blue.flatten()])

    count_zeros = np.sum(lsb_all == 0)
    count_ones = np.sum(lsb_all == 1)

    total = count_zeros + count_ones
    zero_ratio = count_zeros / total
    one_ratio = count_ones / total

    return 1 - abs(zero_ratio - one_ratio)




import warnings
import ctypes
import math
import os

import numpy as np


from stegosphere import base
from stegosphere.config import METADATA_LENGTH_LSB, DELIMITER_MESSAGE
from stegosphere import io
from stegosphere.utils import prng_indices

BACKEND = True
try:
    so_file_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), "backend\\lsb.so")
    backend = ctypes.CDLL(so_file_path)
    backend.embed.argtypes = [ctypes.c_void_p,ctypes.c_int,ctypes.c_char_p,
                              ctypes.c_int,ctypes.c_int]
    backend.embed.restype = None
    backend.extract.argtypes = [ctypes.c_void_p,ctypes.c_int,ctypes.c_int,
                              ctypes.c_int,ctypes.c_char_p]
    backend.extract.restype = None
    
except Exception as e:
    print(e)
    warnings.warn('C backend could not be loaded. Methods will use slower Python.')
    BACKEND = False


def max_capacity(array, bits = 1):
    """
    Calculates the maximum capacity of the cover object for embedding a message.

    The capacity is determined by the size of the data array and the number of bits available for modification.

    :param bits: Number of bits changed per value. Defaults to 1.
    :type bits: int
    :return: The maximum capacity of the object in bits.
    :rtype: int
    """
    return array.size * bits


def embed(array, payload, matching=False, seed=None, bits=1,
               method='metadata', metadata_length=METADATA_LENGTH_LSB,
               delimiter_message=DELIMITER_MESSAGE, compress=False):
    """
    Encodes a message into the cover data using LSB steganography.

    The message can be decoded using either a delimiter, metadata, or without any end-of-message marker. It also
    supports optional message verification to ensure the decoded message matches the original encoded message.
    :param message: The message to be hidden. Gets converted into binary if not already.
    :type message: str
    :param matching: Whether to use LSB matching (not implemented yet). Defaults to False.
    :type matching: bool, optional
    :param seed: (Optional) Seed value for pseudo-randomly distributing the message in the cover data.
    :type seed: int, optional
    :param bits: Number of bits used for decoding per value. Defaults to 1.
    :type bits: int, optional
    :param method: Method for marking the end of the message. Options are 'delimiter', 'metadata', or None. Defaults to 'metadata'.
    :type method: str, optional
    :param metadata_length: Length of the metadata in bits when `method='metadata'`. Defaults to `METADATA_LENGTH_LSB`.
    :type metadata_length: int, optional
    :param delimiter_message: The delimiter string used when `method='delimiter'`.
    :type delimiter_message: str, optional
    :param compress: Whether to use compression on the input data. Defaults to False.
    :type compress: bool, optional

    :raises NotImplementedError: If LSB matching is used.

    :return: True, if it worked.
    :rtype: bool
    """
    if matching is not False: raise NotImplementedError("LSB matching not implemented in current version.")
    
    payload = io.encode_payload(payload, method, metadata_length, delimiter_message, compress)
    
    if len(payload) > max_capacity(array, bits):
        warnings.warn("Insufficient bits, need larger cover or smaller message.")

    orig_dim = array.shape

    if BACKEND is True:
        if seed is not None:
            warnings.warn('set LSB.BACKEND=False when using seed for now.')
        content = array.flatten().copy()
        payload = payload.encode('utf-8')
        value_size = content.dtype.itemsize
        array_pointer = content.ctypes.data_as(ctypes.c_void_p)
        backend.embed(array_pointer, content.size, payload, bits, value_size)
        
    else:
        content = array.copy().flatten().astype(np.int32)
        mask = (1<<bits)-1
        payload_array = np.array([int(payload[i:i + bits], 2) for i in range(0, len(payload), bits)])
        
        payload_length = len(payload_array)
        if seed:
            indices = prng_indices(len(content),seed)[:payload_length]
        else:
            indices = np.arange(payload_length)
        
        content[indices[:payload_length]] &= ~mask
        content[indices[:payload_length]] |= payload_array

    content = content.reshape(orig_dim).astype(array.dtype)

    return content


def extract(array, matching=False, seed=None, bits=1, method='metadata', n_bits=100, 
            metadata_length=METADATA_LENGTH_LSB, delimiter_message=DELIMITER_MESSAGE,
            compress=False):
    """
    Decodes a message from the cover data using LSB steganography.

    :param matching: Whether to use LSB matching (not implemented yet).
    :param seed: Seed value for pseudo-randomly distributing the message in the cover data.
    :param bits: Number of bits used for decoding per value.
    :param method: Method for marking the end of the message: 'delimiter', 'metadata', or None.
    :param n_bits: Bits to be read if method=None.
    :param metadata_length: Length of the metadata in bits if method='metadata'.
    :param delimiter_message: Delimiter used if method='delimiter'.
    :param compress: Whether compression was used on the encoded data.
    :return: The decoded message bits (or a final string if you decompress).
    """
    if matching:
        raise NotImplementedError("LSB matching not implemented in current version.")
    assert method in ['metadata','delimiter', None]
    content = array.copy().flatten()
    
    if seed is not None:
        indices = prng_indices(len(content), seed)
    else:
        indices = np.arange(len(content))


    def extract_bits_python(values, bits):
        mask = (1 << bits) - 1
        return ''.join(f'{(v & mask):0{bits}b}' for v in values)

    def extract_bits_c(values, bits):
        values_flat = values.ravel()
        if not np.issubdtype(values_flat.dtype, np.integer):
            raise ValueError("Array must have an integer dtype for LSB extraction.")
        
        values_cast = values_flat.astype(np.int64, copy=False)

        ptr = values_cast.ctypes.data_as(ctypes.c_void_p)
        length = values_cast.size
        elem_size = values_cast.dtype.itemsize  # e.g. 8 for int64

        total_bits = bits * length
        out_buffer = ctypes.create_string_buffer(total_bits + 1)  # +1 for null terminator
        
        backend.extract(ptr, length, bits, elem_size, out_buffer)
        
        return out_buffer.value.decode('utf-8')

    if BACKEND is True:
        backend_extract = extract_bits_c
    else:
        backend_extract = extract_bits_python


    message_bits = ""

    if method is None:
        if n_bits is None:
            n_bits = len(content) * bits
        
        needed_elems = math.ceil(n_bits / bits)
        subset = content[indices[:needed_elems]]
        
        raw_bits = backend_extract(subset, bits)
        # `raw_bits` could be param * needed_elems in length
        message_bits = raw_bits[:n_bits]

    elif method == 'metadata':
        total_metadata_pixels = math.ceil(metadata_length / bits)
        metadata_subset = content[indices[:total_metadata_pixels]]
        
        metadata_raw = backend_extract(metadata_subset, bits)
        metadata_raw = metadata_raw[:metadata_length]
        message_length = int(metadata_raw, 2)
        
        total_message_pixels = math.ceil(message_length / bits)
        message_subset = content[indices[total_metadata_pixels : total_metadata_pixels + total_message_pixels]]
        
        message_raw = backend_extract(message_subset, bits)
        message_bits = message_raw[:message_length]

    elif method == 'delimiter':
        if not is_binary(delimiter_message):
            delimiter = data_to_binary(delimiter_message)
        else:
            delimiter = delimiter_message
    
        mask = (1 << bits) - 1

        #not efficient for large arrays. replace with reading given amount of values before checking for delimiter
        for idx in indices:
            pixel_val = content[idx]
            extracted_bits = f'{(pixel_val & mask):0{bits}b}'     
            message_bits += extracted_bits
            if message_bits.endswith(delimiter):
                message_bits = message_bits[:-len(delimiter)]  # remove delimiter
                break
    else:
        raise ValueError(f"Invalid method: {method}")

    if compress:
        message_bits = compression.binary_decompress(message_bits, compress)

    return message_bits



import warnings
import math

import numpy as np


from stegosphere import utils
from stegosphere import io
from stegosphere.config import METADATA_LENGTH_VD, DELIMITER_MESSAGE

"""
This is an adapted and generalised version of the Pixel Value Differencing method as proposed by
Wu, D. C., & Tsai, W. H. (2003).
A steganographic method for images by pixel-value differencing.
Pattern recognition letters, 24(9-10), 1613-1626.

"""

def embed(array, payload, spatial_dim=None, channel_dim=None,
          ranges=None, range_offset=3, range_start=1,
          range_range=None,
          seed=None, method='metadata', metadata_length=METADATA_LENGTH_VD,
          delimiter_message=DELIMITER_MESSAGE, compress=False):
            
    """
    Embed a payload into a cover array using Value Differencing steganography.


    :param payload: The payload to be hidden. Gets converted into binary if not already.
    :type payload: str
    :param spatial_dim: Number of spatial dimensions of the data (usually 2 in images, 1 in audio, 3 in video)
    :type spatial_dim: int
    :param channel_dim: Number of channels of the data (3 in RGB, 2 in stereo audio, ...)
    :type channel_dim: int
    :param ranges: pre-defined ranges.
    :type ranges: list
    :param range_offset: ...
    :type range_offset: ...
    :param range_start: ...
    :type range_start: ...
    :param range_range: Minimum and maximum a value can take in the array.
    :type range_range: tuple
    :param seed: (Optional) Seed value for pseudo-randomly distributing the message in the cover data.
    :type seed: int, optional
    :param method: Method for marking the end of the message. Options are 'delimiter', 'metadata', or None. Defaults to 'metadata'.
    :type method: str, optional
    :param metadata_length: Length of the metadata in bits when `method='metadata'`. Defaults to `METADATA_LENGTH_IMAGE`.
    :type metadata_length: int, optional
    :param delimiter_message: The delimiter string used when `method='delimiter'`.
    :type delimiter_message: str, optional

    :return: True, if it worked.
    :rtype: bool
    """
    array = array.copy()
    
    if spatial_dim is None or channel_dim is None:
        if len(array.shape) == 1:
            spatial_dim = 1
            channel_dim = 0
        else:
            spatial_dim = len(array.shape) - 1
            channel_dim = 1
    if ranges is None:
        if range_range is None:
            range_range = utils.dtype_range(array.dtype)
        ranges = _define_range(range_offset, range_start, range_range)

    payload = io.encode_payload(payload, method, metadata_length, delimiter_message, compress)
    value_pairs = _get_pairs(array, spatial_dim, seed)
    payload_index = 0
    max_len = len(payload)
    
    for (v1_coords, v2_coords) in value_pairs:
        if payload_index >= max_len:
            break
        for channel in range(channel_dim):
            v1 = int(array[*v1_coords,channel])
            v2 = int(array[*v2_coords,channel])
            d = v2-v1
            abs_d = abs(d)

            try:
                bits_to_hide, l_i, u_i = _range(ranges, abs_d)
            except Exception:
                continue  # Skip if difference is out of range

            if bits_to_hide>0:
                #Test boundaries:
                if d%2!=0:
                    v1_test = v1 - math.ceil((u_i-d)/2)
                    v2_test = v2 + math.floor((u_i-d)/2)
                if d%2==0:
                    v1_test = v1 - math.floor((u_i-d)/2)
                    v2_test = v2 + math.ceil((u_i-d)/2)

                #if within boundaries
                if 0 <= v1_test <= 255 and 0 <= v2_test <= 255:
                    bits_remaining = max_len - payload_index
                    bits_to_encode = min(bits_to_hide, bits_remaining)
                    bin_secret = payload[payload_index:payload_index + bits_to_encode]
                    if bits_to_encode < bits_to_hide:
                        bin_secret = bin_secret.ljust(bits_to_hide, '0')
                        
                    secret = int(bin_secret,2)
                    
                    new_d = l_i+secret if d >= 0 else -(l_i+secret)
                    
                    if d%2!=0:
                        v1_prime = v1 - math.ceil((new_d-d)/2)
                        v2_prime = v2 + math.floor((new_d-d)/2)
                    if d%2==0:
                        v1_prime = v1 - math.floor((new_d-d)/2)
                        v2_prime = v2 + math.ceil((new_d-d)/2)
                    payload_index += bits_to_hide
                    
                    array[*v1_coords, channel] = v1_prime
                    array[*v2_coords, channel] = v2_prime
                else:
                    
                    continue  # Skip if pixel values go out of bounds

    return array

def extract(array, spatial_dim=None, channel_dim=None,
           ranges=None, range_offset=3, range_start=1, range_range=None,
           seed=None, method='metadata', n_bits=100, metadata_length=METADATA_LENGTH_VD,
           delimiter_message=DELIMITER_MESSAGE, compress=False):
    """
    Decodes a message from the cover data using Value Differencing steganography.

    The message can be decoded using either a delimiter, metadata, or without any end-of-message marker. It also
    supports optional message verification to ensure the decoded message matches the original encoded message.

    :param seed: (Optional) Seed value for pseudo-randomly distributing the message in the cover data.
    :type seed: int, optional
    :param method: Method for marking the end of the message. Options are 'delimiter', 'metadata', or None. Defaults to 'metadata'.
    :type method: str, optional
    :param n_bits: Bits to be read out, used when method=None.
    :type n_bits: int, optional
    :param metadata_length: Length of the metadata in bits when `method='metadata'`. Defaults to `METADATA_LENGTH_LSB`.
    :type metadata_length: int, optional
    :param delimiter_message: The delimiter string used when `method='delimiter'`.
    :type delimiter_message: str, optional

    :return: The decoded message
    :rtype: str or tuple
    """
    assert method in ['metadata','delimiter',None], 'Method must be either delimiter, metadata, or None.'


    if spatial_dim is None or channel_dim is None:
        if len(array.shape) == 1:
            spatial_dim = 1
            channel_dim = 0
        else:
            spatial_dim = len(array.shape) - 1
            channel_dim = 1
    if ranges is None:
        if range_range is None:
            range_range = utils.dtype_range(array.dtype)
        ranges = _define_range(range_offset, range_start, range_range)

        
    bin_payload = ""
    value_pairs = _get_pairs(array, spatial_dim, seed)
    for (v1_coords, v2_coords) in value_pairs:
        for channel in range(channel_dim):
            v1 = int(array[*v1_coords,channel])
            v2 = int(array[*v2_coords,channel])
            d = v2 - v1
            abs_d = abs(d)

            try:
                bits_retrieved, l_i, u_i = _range(ranges, abs_d)
            except Exception:
                continue  # Skip if difference is out of range
            if bits_retrieved > 0:
                #Test boundaries
                if d%2!=0:
                    v1_prime = v1 - math.ceil((u_i-d)/2)
                    v2_prime = v2 + math.floor((u_i-d)/2)
                if d%2==0:
                    v1_prime = v1 - math.floor((u_i-d)/2)
                    v2_prime = v2 + math.ceil((u_i-d)/2)

                if 0 <= v1_prime <= 255 and 0 <= v2_prime <= 255:
                    s = abs_d - l_i
                    bits = bin(s)[2:].zfill(bits_retrieved)
                    bin_payload += bits
                else: 
                    continue  # Skip if pixel values go out of bounds
    
    if method == 'metadata':
        length = int(bin_payload[:metadata_length], 2)
        payload_end = metadata_length + length
        payload = bin_payload[metadata_length:payload_end]
    elif method == 'delimiter':
        if not is_binary(delimiter_message):
            delimiter_message = data_to_binary(delimiter_message)
        payload = bin_payload.split(delimiter_message)[0]
    elif method is None:
        if n_bits is None: 
            return bin_payload
        return bin_payload[:n_bits]

    if compress:
        payload = compression.binary_decompress(payload, compress)
    
    return payload



def _define_range(range_offset, range_start, range_range):
    min_value, max_value = range_range
    max_value += abs(min_value)
    min_value = 0
    lower = 2**range_offset - 1
    ranges = [(range_start,0,lower)]
    lower += 1
    n = 1 + range_offset
    while lower <= max_value:
        upper = min((2**n)-1,max_value)
        ranges.append((n-range_offset+range_start,lower,upper))
        lower = upper + 1
        n += 1       
    return ranges

def _range(ranges, diff):
    """Return the number of bits to embed and the lower and upper bounds based on the pixel difference."""
    #faster then dictionary access
    for i, l_i, u_i in ranges:
        if l_i<=diff<=u_i:
            return i, l_i, u_i
        
def _get_pairs(array, spatial_dim, seed=None):
    #position pairs
    indices = np.ndindex(*array.shape[:spatial_dim])
    indices = list(indices)
    if seed is not None:
        np.random.default_rng(seed).shuffle(indices)
    pairs = [(indices[i], indices[i+1]) for i in range(0, len(indices)-1,2)]
    return pairs



#tool for automated analysis and analysis pipeline

"""
capacity: measures how much data can be stored in a cover
detectability: performs tests how well it can be recognized
efficiency: measures time and memory efficiency
imperceptibility: measures perceivable distortions of cover
robustness: performs attacks
accuracy: measures extraction accuracy

visualizer: visualize differences etc.
"""

from stegosphere.analysis import capacity
from stegosphere.analysis import detectability
from stegosphere.analysis import efficiency
from stegosphere.analysis import imperceptibility
from stegosphere.analysis import robustness
from stegosphere.analysis import accuracy

from stegosphere.analysis import visualizer





import numpy as np
from fontTools.ttLib import TTFont, newTable


def embed(font, payload, table_name):
    """
    Crete new table in ttf file.
    """
    
    if not isinstance(payload, bytes):
        payload = payload.encode('utf-8')
    if table_name in font.font.keys():
        warnings.warn('table name already exists. Data will be overwritten.')
    assert len(table_name)==4, Exception('table tag must be 4 characters')
    custom_table = newTable(table_name)
    custom_table.data = payload
    font.font[table_name] = custom_table
    return font

def extract(font, table_name):
    """
    Extract table from ttf file.
    """
    assert len(table_name)==4, Exception('table tag must be 4 characters')
    if table_name in font.font:
            return font.font[table_name].data
    else:
        #table name not in tables
        return None




from setuptools import setup, find_packages

setup(
    name="stegosphere",
    version="1.2.1",
    description="A flexible steganography and steganalysis library supporting various file types, including encryption and compression",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    author="Maximilian Koch",
    url="https://github.com/Maximilian-Koch/stegosphere",
    packages=find_packages(include=["stegosphere", "stegosphere.*"]),
    include_package_data=True,
    install_requires=[
        "numpy"
    ],
    license="GPL-3.0-or-later",
    python_requires=">=3.6",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
)



from abc import ABC, abstractmethod

class TransformMethod(ABC):
    """
    A superclass for handling transform-based methods.
    """
    @abstractmethod
    def transform(self, **args):
        pass
    @abstractmethod
    def inverse(self, **args):
        pass


class StegMethod(ABC):
    """
    A superclass for handling steganographic methods.
    """
    @abstractmethod
    def embed(self, **args):
        pass
    @abstractmethod
    def extract(self, **args):
        pass





import os

import numpy as np
try:
    from fontTools.ttLib import TTFont, newTable
except:
    raise Exception("Importing fontTools failed. font input needs to be fontTools.TTFont or numpy.array.")


from stegosphere.containers import container

__all__ = ['TTFContainer']

class TTFContainer(container.Container):
    """
    File container for TrueType Font files.
    """
    def __init__(self, font):
        if isinstance(font, TTFont):
            self.font = font
        elif os.path.isfile(font):
            self.font = TTFont(font)
        elif isinstance(font, np.ndarray):
            raise Exception('No need for containers for arrays')
        
    def read(self):
        """
        Read glyphs.
        """
        glyph_data = []
        glyf_table = self.font['glyf']
        glyph_set = self.font.getGlyphSet()
        for glyph_name in glyph_set.keys():
            glyph = glyf_table[glyph_name]
            if hasattr(glyph, 'coordinates') and glyph.coordinates is not None:
                coordinates = glyph.coordinates
                glyph_data.extend(np.array(coordinates).flatten())
        return np.array(glyph_data, dtype=np.int32)
    def flush(self, glyphs):
        """
        Flush new glyphs into container.
        """
        glyf_table = self.font['glyf']
        index = 0
        for glyph_name in self.font.getGlyphSet().keys():
            glyph = glyf_table[glyph_name]
            if not hasattr(glyph, 'coordinates') or glyph.coordinates is None:
                    # Skip composite glyphs and glyphs without coordinates
                continue
            coordinates = glyph.coordinates
            for i in range(len(coordinates)):
                coordinates[i] = (glyphs[index], glyphs[index + 1])
                index += 2
            glyph.coordinates = coordinates
    def save(self, path):
        """
        Save font
        """
        self.font.save(path)



import numpy as np


from stegosphere import utils


def mse(array_1, array_2):
    """
    Calculates the Mean Squared Error (MSE) between the original and encoded array.

    :return: MSE value
    :rtype: float
    """
    return np.mean((array_1.astype(np.float64) - array_2.astype(np.float64)) ** 2)


def psnr(array_1, array_2, max_i=None):
    """
    Calculates the Peak Signal-to-Noise Ratio (PSNR) between the original and encoded array.
    
    :return: PSNR value in dB.
    :rtype: float
    """
    mse_val = mse(array_1, array_2)
    if mse_val == 0:
        return float('inf')
    if max_i is None:
        max_i = utils.dtype_range(array_1.dtype)[1]
    return 10 * np.log10((max_i ** 2) / mse_val)
    

def ssim(array_1, array_2):
    from skimage.metrics import structural_similarity as ss
    return ss(array_1, array_2)


def manhattan_distance(array_1, array_2):
    return np.sum(np.abs(array_1-array_2))

def euclidean_distance(array_1, array_2):
    return np.sqrt(np.sum((array_1-array_2)**2))

def chi_distance(array_1, array_2, epsilon=1e-10):
    p = array_1 / (array_1.sum() + epsilon)
    q = array_2 / (array_2.sum() + epsilon)


def kl_divergence(array_1, array_2, epsilon=1e-10):
    """
    Kullback-Leibler divergence
    """
    p,q = array_1, array_2

    #normalize
    p = p.astype(np.float64)
    q = q.astype(np.float64)
    p /= (p.sum() + epsilon)
    q /= (q.sum() + epsilon)

    kl = np.sum(p * np.log((p + epsilon) / (q + epsilon)))
    return kl


def js_divergence(array_1, array_2):
    """
    Jensen-Shannon divergence
    """
    p,q = array_1, array_2

    #normalize
    p = p.astype(np.float64)
    q = q.astype(np.float64)
    p /= (p.sum() + epsilon)
    q /= (q.sum() + epsilon)
    
    m = 0.5 * (p + q)
    kl_pm = np.sum(p * np.log((p + epsilon) / (m + epsilon)))
    kl_qm = np.sum(q * np.log((q + epsilon) / (m + epsilon)))
    js = 0.5 * (kl_pm + kl_qm)
    return js

def bhattacharyya_divergence(array_1, array_2):
    p,q = array_1, array_2

    #normalize
    p = p.astype(np.float64)
    q = q.astype(np.float64)
    p /= (p.sum() + epsilon)
    q /= (q.sum() + epsilon)
    bc = np.sum(np.sqrt(p*q))
    ba = -np.log(bc+epsilon)
    return ba

def results(array_1, array_2, measures='all', epsilon=1e-10, psnr_max_i=None):
    if measures == 'all' :
        measures = ['mse','psnr','ssim','kl','js','bhattacharyya','manhattan','euclidean','chi']
    #mse, psnr, ssim, kl, js, bhattacharyya, manhattan, euclidean, chi

    output = {}
    if 'psnr' in measures or 'mse' in measures:
        mse = np.mean((self.before.astype(np.float64) - self.after.astype(np.float64)) ** 2)
        if 'mse' in measures:
            output.update({'mse' : mse})
    if 'psnr' in measures:
        if mse == 0:
            psnr = float('inf')
        else:
            if psnr_max_i is None:
                psnr_max_i = utils.dtype_range(self.before.dtype)[1]
            psnr = 10 * np.log10((psnr_max_i ** 2) / mse)
        output.update({'psnr' : psnr})
        
    if 'ssim' in measures:
        pass
    if 'kl' in measures or 'js' in measures or 'bhattacharyya' in measures:
        p,q = array_1, array_2

        #normalize
        p = p.astype(np.float64)
        q = q.astype(np.float64)
        p /= (p.sum() + epsilon)
        q /= (q.sum() + epsilon)
        if 'kl' in measures:
        #Kullback-Leibler divergence
            kl = np.sum(p * np.log((p + epsilon) / (q + epsilon)))
            output.update({'kl':kl})
        if 'js' in measures:
            #Jensen-Shannon divergence
            m = 0.5 * (p + q)
            kl_pm = np.sum(p * np.log((p + epsilon) / (m + epsilon)))
            kl_qm = np.sum(q * np.log((q + epsilon) / (m + epsilon)))
            js = 0.5 * (kl_pm + kl_qm)
            output.update({'js':js})
        if 'bhattacharyya' in measures:
            bc = np.sum(np.sqrt(p*q))
            ba = -np.log(bc+epsilon)
            output.update({'bhattacharyya' : ba})
    if 'manhattan' in measures or 'euclidean' in measures or 'chi' in measures:
        diff = array_1 - array_2
        if 'euclidean' or 'chi' in measures:
            diffsquare = diff**2
            if 'euclidean' in measures:
                euclidean = np.sqrt(np.sum(diffsquare))
                output.update({'euclidean' : euclidean})
            if 'chi' in measures:
                chi = np.sum(diffsquare / (array_1+array_2+epsilon))
                output.update({'chi' : chi})
        if 'manhattan' in measures:
            manhattan = np.sum(np.abs(diff))
            output.update({'manhattan' : manhattan})
            
    return output
